-----------------------------------------------------------------------------------------------------------
--[[
	NaW - Map Class
    Copyright (C) 2011.
	Written by Roland Yonaba - E-mail: roland[dot]yonaba[at]gmail[dot]com
		
		This file holds Map and Minimap Class and metamethods
--]]
-----------------------------------------------------------------------------------------------------------

local ipairs = ipairs
local setmetatable = setmetatable
local floor = math.floor
local ceil = math.ceil
local random = math.random

local Minimap = {}

function Minimap:init(map)
	self.w = 100
	self.h = 100
	self.x = love.graphics.getWidth() - self.w
	self.y = love.graphics.getHeight() - self.h
	self.fovWidth = self.w * (map.backgroundWidth/map.mapWidth)
	self.fovHeight = self.h * (map.backgroundHeight/map.mapHeight)
end

function Minimap:convertCoordinates(map,x,y)
	local x,y = (x / map.mapWidth) * self.w, (y / map.mapHeight) * self.h
	return x,y
end

function Minimap:setCameraInBounds(map,x,y)
	local x,y = self:convertCoordinates(map,x,y)
		if x < 0 then x = 0 end
		if y < 0 then y = 0 end
		if x + self.fovWidth > self.w then x = self.w - self.fovWidth end
		if y + self.fovHeight > self.h then y = self.h - self.fovHeight end
	return x,y
end



function Minimap:draw(map,cursor,...)
	love.graphics.setColor(CONSTANTS.COLORS.CUSTOM)
	love.graphics.rectangle('fill', self.x+1, self.y+1, self.w - 2, self.h - 2)	
	love.graphics.setColor(CONSTANTS.COLORS.YELLOW)
	love.graphics.rectangle('line', self.x, self.y, self.w, self.h)
	
	local x,y = self:setCameraInBounds(map, map.mapX, map.mapY)
	love.graphics.rectangle('line', self.x + x, self.y + y, self.fovWidth, self.fovHeight)
	
	if cursor:isStaticHovering(self) then
		if cursor:isDown('l') and not cursor.isDragging then 
		local mx = cursor.x - self.x
		local my = cursor.y - self.y
		map.mapX = (mx/self.w) * map.mapWidth - map.backgroundWidth/2
		map.mapY  = (my/self.h) * map.mapHeight - map.backgroundHeight/2
		end
	end
	
	
	--love.graphics.setColor(CONSTANTS.COLORS.GREEN)
	for k,t_arg in ipairs(arg) do
	love.graphics.setColor(CONSTANTS.COLORS[t_arg.color])
		for i,item in ipairs(t_arg.items) do
			local x,y = self:convertCoordinates(map,item.x,item.y)			
			love.graphics.rectangle('fill',self.x + x, self.y + y, 2, 2)
		end
	end
	love.graphics.setColor(CONSTANTS.COLORS.WHITE)
end

setmetatable(Minimap, {__call = function(self,map) return self:init(map) end})

local Map = {} 

function Map:init(width,height,tileWidth,tileHeight)
	self.cMap = {}
		for y = 1,floor(height/tileHeight) do self.cMap[y] = {}
			for x = 1,floor(width/tileWidth) do self.cMap[y][x] = 0 end
		end
	self.mapHeight = height
	self.mapWidth = width
	self.tileHeight = tileHeight
	self.tileWidth = tileWidth
	self.cMapHeight = #self.cMap
	self.cMapWidth = #self.cMap[1]
	self.mapX = 0
	self.mapY = 0
	self.SCROLL = 1	
end


function Map:setBackground(im)
	self.background = love.graphics.newImage(im)
	self.backgroundWidth = self.background:getWidth()
	self.backgroundHeight = self.background:getHeight()
	self.n_x_background = floor(self.mapWidth/self.backgroundWidth)
	self.n_y_background = floor(self.mapHeight/self.backgroundHeight)
	self.minimap = Minimap
	self.minimap(self)
	self.terrain = { chasms = random(1,3), mountains = random(1,3), locations = { x_s = {}, y_s={} } }	
	self.terrain.img = {
							chasm = love.graphics.newImage('Images/Resources/Chasm.png'),
							mountain = love.graphics.newImage('Images/Resources/Mountain.png'),
						}
	self.terrain.size = { 
							chasm = { width = self.terrain.img.chasm:getWidth(), height = self.terrain.img.chasm:getHeight() },
							mountain = { width = self.terrain.img.mountain:getWidth(), height = self.terrain.img.mountain:getHeight() },
						}	
	
	local x, y, w, h
		for i = 1, self.terrain.chasms do
			self.terrain.locations.x_s[i] = random(self.mapWidth)
			self.terrain.locations.y_s[i] = random(self.mapHeight)
			self.terrain.locations[i] = 'chasm'			
			x = floor(self.terrain.locations.x_s[i]/self.tileWidth) + 1
			y = floor(self.terrain.locations.y_s[i]/self.tileHeight) + 1
			w = ceil(self.terrain.size.chasm.width/self.tileWidth)
			h = ceil(self.terrain.size.chasm.height/self.tileHeight)
				for cy = y, y + h do
					for cx = x, x + w do
						if self.cMap[cy] and self.cMap[cy][cx] then self.cMap[cy][cx] = 1 end
					end
				end				
		end
		for i = self.terrain.chasms + 1, self.terrain.chasms + self.terrain.mountains do
			self.terrain.locations.x_s[i] = random(self.mapWidth)
			self.terrain.locations.y_s[i] = random(self.mapHeight)
			self.terrain.locations[i] = 'mountain'
			x = floor(self.terrain.locations.x_s[i]/self.tileWidth) + 1
			y = floor(self.terrain.locations.y_s[i]/self.tileHeight) + 1
			w = ceil(self.terrain.size.mountain.width/self.tileWidth)
			h = ceil(self.terrain.size.mountain.height/self.tileHeight)
				for cy = y, y + h do
					for cx = x, x + w do
						if self.cMap[cy] and self.cMap[cy][cx] then self.cMap[cy][cx] = 1 end
					end
				end				
		end		

end

function Map:resetView()
	if self.mapX < 0 then self.mapX = 0 end
	if self.mapY < 0 then self.mapY = 0 end
	if self.mapX > self.mapWidth then self.mapX = self.mapWidth end
	if self.mapY > self.mapHeight then self.mapY = self.mapHeight end
end

function Map:update(delta_time, cursor)
	if love.keyboard.isDown('left') or cursor.x <= 0 then 
		if self.mapX > 0 then 
			self.mapX = self.mapX - self.SCROLL * delta_time			
		else self.mapX=0 end
	end
	if love.keyboard.isDown('right') or cursor.x >= self.backgroundWidth-cursor.w then
		if self.mapX < self.mapWidth - self.backgroundWidth then 
			self.mapX = self.mapX + self.SCROLL * delta_time 			
		else self.mapX = self.mapWidth - self.backgroundWidth end
	end
	if love.keyboard.isDown('up') or cursor.y <= 0 then
		if self.mapY > 0 then 
			self.mapY = self.mapY - self.SCROLL * delta_time 
		else self.mapY = 0 end
	end
	if love.keyboard.isDown('down') or cursor.y >= self.backgroundHeight-cursor.h then
		if self.mapY < self.mapHeight - self.backgroundHeight then 
			self.mapY = self.mapY + self.SCROLL * delta_time
		else self.mapY = self.mapHeight - self.backgroundHeight end
	end
	self:resetView()
end

function Map:writeCMap()
	local f = love.filesystem.newFile('Maps/cMap.txt')
	if f then
	f:open('w')
		for y in ipairs(self.cMap) do
			for x in ipairs(self.cMap[y]) do 
			f:write(self.cMap[y][x])		
			end
		f:write('\n')
		end
	f:close()		
	end
end 

function Map:draw()	
	local xint = - self.backgroundWidth
	local yint = 0
	for y = 1,self.n_y_background+1 do
		for x = 1,self.n_x_background+1 do
		xint = xint + self.backgroundWidth
			if (self.mapX < xint + self.backgroundWidth) and 
				(self.mapX + self.backgroundWidth > xint) and (self.mapY < yint + self.backgroundHeight) and 
				(self.mapY + self.backgroundHeight  > yint) then
			love.graphics.draw(self.background,xint - self.mapX,yint - self.mapY)
			end
		end
	yint = yint + self.backgroundHeight 
	xint = - self.backgroundWidth	
	end		
	for i = 1, (#self.terrain.locations) do
	local terrain_type = self.terrain.locations[i]
		if ((self.mapX < self.terrain.locations.x_s[i] + self.terrain.size[terrain_type].width) and 
			(self.mapX + self.backgroundWidth > self.terrain.locations.x_s[i]) and 
			(self.mapY < self.terrain.locations.y_s[i] + self.terrain.size[terrain_type].height) and 
			(self.mapY + self.backgroundHeight > self.terrain.locations.y_s[i])) then
		love.graphics.draw( self.terrain.img[terrain_type], self.terrain.locations.x_s[i] - self.mapX, self.terrain.locations.y_s[i] - self.mapY)
		end
	end
end

return setmetatable(Map, {__call = function(self,...) return self:init(...) end})







