-----------------------------------------------------------------------------------------------------------
--[[
	NaW - Units Manager Class
    Copyright (C) 2011.
	Written by Roland Yonaba - E-mail: roland[dot]yonaba[at]gmail[dot]com
		
		This file Units behaviour
--]]
-----------------------------------------------------------------------------------------------------------

local require  = require
local ipairs = ipairs
local ceil = math.ceil
local floor = math.floor
local abs = math.abs
local insert = table.insert
local remove = table.remove

local CONSTANTS = require 'Scripts/Constants'
local Utils = require 'Scripts/Utils/Unit'
local Astar = require 'Scripts/Libraries/Astar'


local pathfinding_queue = {}
local Images = Utils.loadAllUnits('Images/Units/')

local Unit =  { x = 0, y = 0, w = 0, h = 0, health = 0, timer = 0, type = '',
		  move = false, selected = false, speed = 0, state = 'Standing', direction = 1,
		  path = {}, cur = 0, pathLen = 0, temp_x = 0, temp_y = 0,
		  img = ''}

Unit.all = {}

function Unit:initPathfinding(map, obst, diagonal, depth)
	Astar(map, obst)
	Astar.diagonalMove = diagonal or false
	Astar:setSearchDepth(depth or 0)
end

function Unit:resetPathfinding(obst)
	Astar:reset(obst)
end

function Unit:new(x, y, utype)
	local u = {}
	self.__index = self
	setmetatable(u,self)
	u:setAbsolutePosition(x or 0, y or 0)	
	u.type = utype
	u.health = self.Parameters[u.type].Health
	u.speed = self.Parameters[u.type].Speed
	u.img = Images[u.type][u.state][u.direction][1]
	u.w = u.img:getWidth()
	u.h = u.img:getHeight()
	insert(self.all,u)
end
		  

Unit.Parameters = { 
					Villager = {Health = 100, Speed = 1},
				  }

function Unit:setAbsolutePosition(x,y)
	self.x = x
	self.y = y
end

function Unit:setMapPosition(map,x,y)
	self.x = map.tileWidth * x - map.tileWidth
	self.y = map.tileHeight * y - map.tileHeight
end

function Unit:isVisible(map)	
	if  (self.health > 0) and ((map.mapX < self.x + self.w) and 
		(map.mapX + map.backgroundWidth > self.x) and 
		(map.mapY < self.y+self.h) and 
		(map.mapY + map.backgroundHeight > self.y)) then
		return true
	end
	return false
end

function Unit:update(delta_time, map, cursor)
	for i,unit in ipairs(self.all) do
	unit.mX = floor(unit.x / map.tileWidth)+1
	unit.mY = floor(unit.y / map.tileHeight)+1
		if unit.state == 'Standing' then
		unit.timer = 0
		unit.img = Images[unit.type].Standing[unit.direction][1]
		elseif unit.state == 'Moving' then
			if unit.x < unit.temp_x then
			unit.x = unit.x + unit.speed * (delta_time)
			elseif unit.x > unit.temp_x then
			unit.x = unit.x - unit.speed * (delta_time)
			end
			if unit.y < unit.temp_y then
			unit.y = unit.y + unit.speed * (delta_time)
			elseif unit.y > unit.temp_y then
			unit.y = unit.y - unit.speed * (delta_time)
			end				
			if (abs(unit.x - unit.temp_x) <= 2) and (abs(unit.y-unit.temp_y) <= 2) then
				if unit.cur < unit.pathLen then
				unit.cur = unit.cur + 1
				unit.temp_x = unit.path[unit.cur].x * map.tileWidth - map.tileWidth
				unit.temp_y = unit.path[unit.cur].y * map.tileHeight - map.tileHeight
				else
				unit.state = 'Standing'
				unit.x = unit.path[unit.pathLen].x * map.tileWidth - map.tileWidth
				unit.y = unit.path[unit.pathLen].y * map.tileHeight - map.tileHeight
				unit.cur = 0
				unit.pathLen = 0
				unit.path = {}
				unit.temp_x,unit.temp_y = 0,0
				end
			end		
		end		
		if unit.selected then
			if cursor:pressedOnce('l') then
			local sNode = { x = unit.mX, y = unit.mY }
			local fNode = { x = floor((cursor.x + map.mapX) / map.tileWidth)+1, y = floor((cursor.y + map.mapY) / map.tileHeight)+1 }
				if sNode.x ~= fNode.x and sNode.y~=fNode.y then
				insert(pathfinding_queue,{id = i, start = sNode, final = fNode})
				print('start: x '..sNode.x..' y: '..sNode.y..' End : x '..fNode.x..' y: '..fNode.y)
				end
			end
		end
	end
	--[
	--]]
end

function Unit:drawAll(map,cursor)
	if #pathfinding_queue > 0 then
	Astar:setInitialNode(pathfinding_queue[1].start.x, pathfinding_queue[1].start.y)
	Astar:setFinalNode(pathfinding_queue[1].final.x, pathfinding_queue[1].final.y)
	print('start computing',love.timer.getTime())
	local path = Astar:getPath()
	print('end computing',love.timer.getTime())
		if path then
		local i = pathfinding_queue[1].id
		self.all[i].path = path
		self.all[i].state = 'Moving'
		self.all[i].pathLen = #path
		self.all[i].cur = 1
		self.all[i].temp_x = self.all[i].path[1].x * map.tileWidth - map.tileWidth
		self.all[i].temp_y = self.all[i].path[1].y * map.tileHeight - map.tileHeight
		else
		Astar:reset(1)
		end
	remove(pathfinding_queue,1)
	end

	for i,unit in ipairs(self.all) do
		if cursor:dragging(unit) then
		unit.selected = true
		end
		if unit:isVisible(map) then
			if cursor:isHovering(unit, map.mapX, map.mapY) and cursor:isDown('l') then
			unit.selected = true
			end
			if unit.selected then
			love.graphics.setColor(CONSTANTS.COLORS.RED)
			love.graphics.rectangle('fill',unit.x - map.mapX, unit.y - 10 - map.mapY, 30, 2)
			love.graphics.setColor(CONSTANTS.COLORS.GREEN)
			love.graphics.rectangle('fill',unit.x - map.mapX, unit.y - 10 - map.mapY, 30 * ceil(unit.health/self.Parameters[unit.type].Health), 2)
			love.graphics.setColor(CONSTANTS.COLORS.WHITE)			
			end
			love.graphics.draw(unit.img, unit.x - map.mapX, unit.y-map.mapY)
		end
	end
end

function Unit:deselectAll(button)
	if button == 'r' then
		for k in ipairs(self.all) do 
		self.all[k].selected = false 
		end
	end
end

return setmetatable(Unit, {__call = function(self,...) return self:new(...) end} )